//------------------------------------------------------------------------
// Tester for the 3D Math Library.
// Copyright (C) 2001-2002 David Poon
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General 
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//------------------------------------------------------------------------

#include <cstdio>
#include "mathlib.hpp"

bool TestCastToFloatPtr(float *array = 0)
{
	if (!array)
		return false;

	return true;
}

void TestMatrix()
{
	/*
	Expected Output for TestMatrix():
	----------------------------------------------
	Testing Matrix portion of Maths Library...
	Test #1: Matrix float* cast operator PASSED
	Test #2: Matrix (row, col) operator PASSED
	Test #3: Matrix multiplication operator PASSED
	Test #4: Matrix transpose PASSED
	Test #5: Matrix inverse PASSED
	Test #6: Matrix addition PASSED
	Test #7: Matrix subtraction PASSED
	Test #8: Matrix multiplication by scalar PASSED
	Test #9: (1.f / 10.f) * 10.f == 1.f  PASSED
	Test #10: Build matrix to rotate from one vector to another
			Rotate(from, to) Case 1 PASSED
			Rotate(from, to) Case 2 PASSED
			Rotate(from, to) Case 3 PASSED
	Test #11: Build translation matrix PASSED
	Test #12: Build scaling matrix PASSED
	Test #13: Build rotation matrix PASSED
	*/

	puts("\nTesting Matrix portion of Maths Library...");

	// Test #1: Matrix float* cast operator.
	{
		Matrix m;
		if (!TestCastToFloatPtr(m))
			puts("Test #1: Matrix float* cast operator FAILED");
		else
			puts("Test #1: Matrix float* cast operator PASSED");
	}

	// Test #2: Matrix (row, col) operator.
	{
		Matrix m;
		m(1, 2) = 12.f;
		if (!CloseEnough(m(1, 2), 12.f))
			puts("Test #2: Matrix (row, col) operator FAILED");
		else
			puts("Test #2: Matrix (row, col) operator PASSED");
	}

	// Test #3: Matrix multiplication operator.
	{
		Matrix m;
		m(0,0) = 1.f;	m(0,1) = 1.f;	m(0,2) = 1.f;	m(0,3) = 1.f;
		m(1,0) = 2.f;	m(1,1) = 2.f;	m(1,2) = 2.f;	m(1,3) = 2.f;
		m(2,0) = 3.f;	m(2,1) = 3.f;	m(2,2) = 3.f;	m(2,3) = 3.f;
		m(3,0) = 4.f;	m(3,1) = 4.f;	m(3,2) = 4.f;	m(3,3) = 4.f;

		Matrix identity, result;
		
		identity.loadIdentity();
		result.multiply(m, identity);

		if (result != m)
			puts("Test #3: Matrix multiplication operator FAILED");
		else
			puts("Test #3: Matrix multiplication operator PASSED");
	}

	// Test #4: Matrix transpose.
	{
		Matrix m; // original matrix
		m(0,0) = 1.f;	m(0,1) = 1.f;	m(0,2) = 1.f;	m(0,3) = 1.f;
		m(1,0) = 2.f;	m(1,1) = 2.f;	m(1,2) = 2.f;	m(1,3) = 2.f;
		m(2,0) = 3.f;	m(2,1) = 3.f;	m(2,2) = 3.f;	m(2,3) = 3.f;
		m(3,0) = 4.f;	m(3,1) = 4.f;	m(3,2) = 4.f;	m(3,3) = 4.f;
		
		Matrix t; // the expected transpose of m
		t(0,0) = 1.f;	t(0,1) = 2.f;	t(0,2) = 3.f;	t(0,3) = 4.f;
		t(1,0) = 1.f;	t(1,1) = 2.f;	t(1,2) = 3.f;	t(1,3) = 4.f;
		t(2,0) = 1.f;	t(2,1) = 2.f;	t(2,2) = 3.f;	t(2,3) = 4.f;
		t(3,0) = 1.f;	t(3,1) = 2.f;	t(3,2) = 3.f;	t(3,3) = 4.f;

		Matrix result;
		result.transpose(m);

		if (result != t)
			puts("Test #4: Matrix transpose FAILED");
		else
			puts("Test #4: Matrix transpose PASSED");
	}

	// Test #5: Matrix inverse.
	{
		Matrix identity; // identity matrix
		identity.loadIdentity();

		Matrix m; // original matrix.
		m(0,0) = 1.f;	m(0,1) = -3.f;	m(0,2) = 0.f;	m(0,3) = -2.f;
		m(1,0) = 3.f;	m(1,1) = -12.f;	m(1,2) = -2.f;	m(1,3) = -6.f;
		m(2,0) = -2.f;	m(2,1) = 10.f;	m(2,2) = 2.f;	m(2,3) = 5.f;
		m(3,0) = -1.f;	m(3,1) = 6.f;	m(3,2) = 1.f;	m(3,3) = 3.f;

		Matrix i; // the expected inverse of m
		i(0,0) = 0.f;	i(0,1) = 1.f;	i(0,2) = 0.f;	i(0,3) = 2.f;
		i(1,0) = 1.f;	i(1,1) = -1.f;	i(1,2) = -2.f;	i(1,3) = 2.f;
		i(2,0) = 0.f;	i(2,1) = 1.f;	i(2,2) = 3.f;	i(2,3) = -3.f;
		i(3,0) = -2.f;	i(3,1) = 2.f;	i(3,2) = 3.f;	i(3,3) = -2.f;

		Matrix inverse;
		inverse.inverse(m);
		
		if (inverse == i)
		{
			inverse.multiply(m);

			if (inverse == identity)
				puts("Test #5: Matrix inverse PASSED");
			else
				puts("Test #5: Matrix inverse FAILED");
		}
		else
		{
			puts("Test #5: Matrix inverse FAILED");
		}
	}

	// Test #6: Matrix addition.
	{
		Matrix m1, m2;
		m1.loadIdentity();
		m2.loadIdentity();
		
		Matrix m3; // expected result
		m3(0,0) = 2.f;	m3(0,1) = 0.f;	m3(0,2) = 0.f;	m3(0,3) = 0.f;
		m3(1,0) = 0.f;	m3(1,1) = 2.f;	m3(1,2) = 0.f;	m3(1,3) = 0.f;
		m3(2,0) = 0.f;	m3(2,1) = 0.f;	m3(2,2) = 2.f;	m3(2,3) = 0.f;
		m3(3,0) = 0.f;	m3(3,1) = 0.f;	m3(3,2) = 0.f;	m3(3,3) = 2.f;

		m1.add(m2);

		if (m1 != m3)
			puts("Test #6: Matrix addition FAILED");
		else
			puts("Test #6: Matrix addition PASSED");
	}

	// Test #7: Matrix subtraction.
	{
		Matrix m1, m2; 
		m1.loadIdentity();
		m2.loadIdentity();

		Matrix m3; // expected result
		m3(0,0) = 0.f;	m3(0,1) = 0.f;	m3(0,2) = 0.f;	m3(0,3) = 0.f;
		m3(1,0) = 0.f;	m3(1,1) = 0.f;	m3(1,2) = 0.f;	m3(1,3) = 0.f;
		m3(2,0) = 0.f;	m3(2,1) = 0.f;	m3(2,2) = 0.f;	m3(2,3) = 0.f;
		m3(3,0) = 0.f;	m3(3,1) = 0.f;	m3(3,2) = 0.f;	m3(3,3) = 0.f;

		m1.subtract(m2);

		if (m1 != m3)
			puts("Test #7: Matrix subtraction FAILED");
		else
			puts("Test #7: Matrix subtraction PASSED");
	}

	// Test #8: Matrix multiplication by scalar.
	{
		Matrix m1; // original matrix
		m1(0,0) = 1.f;	m1(0,1) = 1.f;	m1(0,2) = 1.f;	m1(0,3) = 1.f;
		m1(1,0) = 1.f;	m1(1,1) = 1.f;	m1(1,2) = 1.f;	m1(1,3) = 1.f;
		m1(2,0) = 1.f;	m1(2,1) = 1.f;	m1(2,2) = 1.f;	m1(2,3) = 1.f;
		m1(3,0) = 1.f;	m1(3,1) = 1.f;	m1(3,2) = 1.f;	m1(3,3) = 1.f;

		Matrix m2; // expected result
		m2(0,0) = 3.f;	m2(0,1) = 3.f;	m2(0,2) = 3.f;	m2(0,3) = 3.f;
		m2(1,0) = 3.f;	m2(1,1) = 3.f;	m2(1,2) = 3.f;	m2(1,3) = 3.f;
		m2(2,0) = 3.f;	m2(2,1) = 3.f;	m2(2,2) = 3.f;	m2(2,3) = 3.f;
		m2(3,0) = 3.f;	m2(3,1) = 3.f;	m2(3,2) = 3.f;	m2(3,3) = 3.f;

		m1.multiply(3.f);

		if (m1 == m2)
			puts("Test #8: Matrix multiplication by scalar PASSED");
		else
			puts("Test #8: Matrix multiplication by scalar FAILED");
	}

	// Test #9: Comparing floating points for equality.
	{
		float tenth = 1.f / 10.f;

		if (tenth * 10.f != 1.f)
		{
			if (CloseEnough(tenth * 10.f, 1.f))
				puts("Test #9: (1.f / 10.f) * 10.f == 1.f  PASSED");
		}
		else
		{
			puts("Test #9: (1.f / 10.f) * 10.f == 1.f  FAILED");
		}
	}

	// Test #10: Rotation of vector FROM into vector TO.
	{
		puts("Test #10: Build matrix to rotate from one vector to another");

		Vector temp;
		Vector from;
		Vector to;
		Matrix m;

		// Case 1: from = (1,0,0), to = (1,0,0)
		from.set(1.f, 0.f, 0.f);
		to = from;
		
		m.rotate(from, to);
		temp.multiply(from, m);

		if (temp == to)
			puts("\tRotate(from, to) Case 1 PASSED");
		else
			puts("\tRotate(from, to) Case 1 FAILED");

		// Case 2: from = (1,0,0), to = (-1,0,0)
		from.set(1.f, 0.f, 0.f);
		to.set(-1.f, 0.f, 0.f);
		
		m.rotate(from, to);
		temp.multiply(from, m);
		
		if (temp == to)
			puts("\tRotate(from, to) Case 2 PASSED");
		else
			puts("\tRotate(from, to) Case 2 FAILED");

		// Case 3: from = (1,0,0), to = (0,1,0)
		from.set(1.f, 0.f, 0.f);
		to.set(0.f, 1.f, 0.f);
		
		m.rotate(from, to);
		temp.multiply(from, m);
		
		if (temp == to)
			puts("\tRotate(from, to) Case 3 PASSED");
		else
			puts("\tRotate(from, to) Case 3 FAILED");
	}

	// Test #11: Build translation matrix.
	{
		Point p(0.f, 0.f, 0.f);
		Point expected(7.f, 7.f, 7.f);

		Matrix m;
		m.translate(7.f, 7.f, 7.f);

		Point result;
		result.multiply(p, m);

		if (result == expected)
			puts("Test #11: Build translation matrix PASSED");
		else
			puts("Test #11: Build translation matrix FAILED");
	}

	// Test #12: Build scaling matrix.
	{
		Point p(1.f, 1.f, 1.f);
		Point expected(7.f, 7.f, 7.f);

		Matrix m;
		m.scale(7.f, 7.f, 7.f);

		Point result;
		result.multiply(p, m);

		if (result == expected)
			puts("Test #12: Build scaling matrix PASSED");
		else
			puts("Test #12: Build scaling matrix FAILED");
	}

	// Test #13: Build rotation matrix.
	{
		Point p(1.f, 0.f, 0.f);
		Point expected(0.f, 1.f, 0.f);

		Matrix m;
		Vector zAxis(0.f, 0.f, 1.f); // rotate about z-axis
		m.rotate(90.f, zAxis);

		Point result;
		result.multiply(p, m);

		if (result == expected)
			puts("Test #13: Build rotation matrix PASSED");
		else
			puts("Test #13: Build rotation matrix FAILED");
	}
}

void TestPoint()
{
	/*
	Expected Output for TestPoint():
	----------------------------------------------
	Testing Point portion of Maths Library...
	Test #1: Translate Point PASSED
	Test #2: Scale Point PASSED
	Test #3: Rotate Point PASSED
	*/

	puts("\nTesting Point portion of Maths Library...");

	// Test #1: Translate Point.
	{
		Point p(0.f, 0.f, 0.f);
		Point expected(7.f, 7.f, 7.f);

		Matrix m;
		m.translate(7.f, 7.f, 7.f);

		Point result;
		result.multiply(p, m);

		if (result == expected)
			puts("Test #1: Translate Point PASSED");
		else
			puts("Test #1: Translate Point FAILED");
	}

	// Test #2: Scale Point.
	{
		Point p(1.f, 1.f, 1.f);
		Point expected(7.f, 7.f, 7.f);

		Matrix m;
		m.scale(7.f, 7.f, 7.f);

		Point result;
		result.multiply(p, m);

		if (result == expected)
			puts("Test #2: Scale Point PASSED");
		else
			puts("Test #2: Scale Point FAILED");
	}

	// Test #3: Rotate Point.
	{
		Point p(1.f, 0.f, 0.f);
		Point expected(0.f, 1.f, 0.f);

		Matrix m;
		Vector zAxis(0.f, 0.f, 1.f); // rotate about z-axis
		m.rotate(90.f, zAxis);

		Point result;
		result.multiply(p, m);

		if (result == expected)
			puts("Test #3: Rotate Point PASSED");
		else
			puts("Test #3: Rotate Point FAILED");
	}
}

void TestVector()
{
	/*
	Expected Output for TestVector():
	----------------------------------------------
	*/

	puts("\nTesting Vector portion of Maths Library...");
}

void TestQuaternion()
{
	/*
	Expected Output for TestQuaternion():
	----------------------------------------------
	*/

	puts("\nTesting Quaternion portion of Maths Library...");
}

void TestPlane()
{
	/*
	Expected Output for TestPlane():
	----------------------------------------------
	Planes PASSED
	*/

	puts("\nTesting Plane portion of Maths Library...");

	Point points[] =
	{
		Point(-1.f, 1.f, 1.f),
		Point(1.f,  1.f, 1.f),
		Point(1.f, -1.f, 1.f)
	};

	Plane plane(points[0], points[1], points[2]);
	Point infront(0.f, 0.f, 0.5f);
	Point coplanar(1.f, 1.f, 1.f);
	Point behind(0.f, 0.f, 5.f);

	if (plane.distanceTo(infront) > 0.f 
		&& plane.distanceTo(coplanar) == 0.f
		&& plane.distanceTo(behind) < 0.f)
	{
		puts("Planes PASSED");
	}
	else
	{
		puts("Planes FAILED");
	}
}

int main()
{
	TestMatrix();
	TestPoint();
	TestVector();
	TestQuaternion();
	TestPlane();
	return 0;
}